"""One shared terminal per URL endpoint

Plus a /new URL which will create a new terminal and redirect to it.
"""

import tornado.web

# This demo requires tornado_xstatic and XStatic-term.js
import tornado_xstatic
from common_demo_stuff import STATIC_DIR, TEMPLATE_DIR, run_and_show_browser

from terminado import NamedTermManager, TermSocket

AUTH_TYPES = ("none", "login")


class TerminalPageHandler(tornado.web.RequestHandler):
    """Render the /ttyX pages"""

    def get(self, term_name):
        return self.render(
            "termpage.html",
            static=self.static_url,
            xstatic=self.application.settings["xstatic_url"],
            ws_url_path="/_websocket/" + term_name,
        )


class NewTerminalHandler(tornado.web.RequestHandler):
    """Redirect to an unused terminal name"""

    def get(self):
        name, terminal = self.application.settings["term_manager"].new_named_terminal()
        self.redirect("/" + name, permanent=False)


def main():
    term_manager = NamedTermManager(shell_command=["bash"], max_terminals=100)

    handlers = [
        (r"/_websocket/(\w+)", TermSocket, {"term_manager": term_manager}),
        (r"/new/?", NewTerminalHandler),
        (r"/(\w+)/?", TerminalPageHandler),
        (r"/xstatic/(.*)", tornado_xstatic.XStaticFileHandler),
    ]
    application = tornado.web.Application(
        handlers,
        static_path=STATIC_DIR,
        template_path=TEMPLATE_DIR,
        xstatic_url=tornado_xstatic.url_maker("/xstatic/"),
        term_manager=term_manager,
    )

    application.listen(8700, "localhost")
    run_and_show_browser("http://localhost:8700/new", term_manager)


if __name__ == "__main__":
    main()



"""Tornado websocket handler to serve a terminal interface.
"""
# Copyright (c) Jupyter Development Team
# Copyright (c) 2014, Ramalingam Saravanan <sarava@sarava.net>
# Distributed under the terms of the Simplified BSD License.
from __future__ import annotations

import json
import logging
import os
from typing import TYPE_CHECKING, Any

import tornado.websocket
from tornado import gen
from tornado.concurrent import run_on_executor

if TYPE_CHECKING:
    from terminado.management import PtyWithClients, TermManagerBase


def _cast_unicode(s: str | bytes) -> str:
    if isinstance(s, bytes):
        return s.decode("utf-8")
    return s


class TermSocket(tornado.websocket.WebSocketHandler):
    """Handler for a terminal websocket"""

    def initialize(self, term_manager: TermManagerBase) -> None:
        """Initialize the handler."""
        self.term_manager = term_manager
        self.term_name = ""
        self.size = (None, None)
        self.terminal: PtyWithClients | None = None
        self._blocking_io_executor = term_manager.blocking_io_executor

        self._logger = logging.getLogger(__name__)
        self._user_command = ""

        # Enable if the environment variable LOG_TERMINAL_OUTPUT is "true"
        self._enable_output_logging = str.lower(os.getenv("LOG_TERMINAL_OUTPUT", "false")) == "true"

    def origin_check(self, origin: str | None = None) -> bool:
        """Deprecated: backward-compat for terminado <= 0.5."""
        origin = origin or self.request.headers.get("Origin", "")
        assert origin is not None
        return self.check_origin(origin)

    def open(self, url_component: Any = None) -> None:  # type:ignore[override]
        """Websocket connection opened.

        Call our terminal manager to get a terminal, and connect to it as a
        client.
        """
        # Jupyter has a mixin to ping websockets and keep connections through
        # proxies alive. Call super() to allow that to set up:
        super().open(url_component)

        self._logger.info("TermSocket.open: %s", url_component)

        url_component = _cast_unicode(url_component)
        self.term_name = url_component or "tty"
        self.terminal = self.term_manager.get_terminal(url_component)
        self.terminal.clients.append(self)
        self.send_json_message(["setup", {}])
        self._logger.info("TermSocket.open: Opened %s", self.term_name)
        # Now drain the preopen buffer, if reconnect.
        buffered = ""
        preopen_buffer = self.terminal.read_buffer.copy()
        while True:
            if not preopen_buffer:
                break
            s = preopen_buffer.popleft()
            buffered += s
        if buffered:
            self.on_pty_read(buffered)

    def on_pty_read(self, text: str) -> None:
        """Data read from pty; send to frontend"""
        self.send_json_message(["stdout", text])

    def send_json_message(self, content: Any) -> None:
        """Send a json message on the socket."""
        json_msg = json.dumps(content)
        self.write_message(json_msg)

        if self._enable_output_logging and content[0] == "stdout" and isinstance(content[1], str):
            self.log_terminal_output(f"STDOUT: {content[1]}")

    @gen.coroutine
    def on_message(self, message: str) -> None:  # type:ignore[misc]
        """Handle incoming websocket message

        We send JSON arrays, where the first element is a string indicating
        what kind of message this is. Data associated with the message follows.
        """
        # logging.info("TermSocket.on_message: %s - (%s) %s", self.term_name, type(message), len(message) if isinstance(message, bytes) else message[:250])
        command = json.loads(message)
        msg_type = command[0]
        assert self.terminal is not None
        if msg_type == "stdin":
            yield self.stdin_to_ptyproc(command[1])
            if self._enable_output_logging:
                if command[1] == "\r":
                    self.log_terminal_output(f"STDIN: {self._user_command}")
                    self._user_command = ""
                else:
                    self._user_command += command[1]
        elif msg_type == "set_size":
            self.size = command[1:3]
            self.terminal.resize_to_smallest()

    def on_close(self) -> None:
        """Handle websocket closing.

        Disconnect from our terminal, and tell the terminal manager we're
        disconnecting.
        """
        self._logger.info("Websocket closed")
        if self.terminal:
            self.terminal.clients.remove(self)
            self.terminal.resize_to_smallest()
        self.term_manager.client_disconnected(self)

    def on_pty_died(self) -> None:
        """Terminal closed: tell the frontend, and close the socket."""
        self.send_json_message(["disconnect", 1])
        self.close()
        self.terminal = None

    def log_terminal_output(self, log: str = "") -> None:
        """
        Logs the terminal input/output
        :param log: log line to write
        :return:
        """
        self._logger.debug(log)

    @run_on_executor(executor="_blocking_io_executor")
    def stdin_to_ptyproc(self, text: str) -> None:
        """Handles stdin messages sent on the websocket.

        This is a blocking call that should NOT be performed inside the
        server primary event loop thread. Messages must be handled
        asynchronously to prevent blocking on the PTY buffer.
        """
        if self.terminal is not None:
            self.terminal.ptyproc.write(text)



import webbrowser
from pathlib import Path

import tornado.ioloop

import terminado

HERE = Path(terminado.__file__).parent
STATIC_DIR = HERE / "_static"
TEMPLATE_DIR = HERE / "templates"


def run_and_show_browser(url, term_manager):
    loop = tornado.ioloop.IOLoop.instance()
    loop.add_callback(webbrowser.open, url)
    try:
        loop.start()
    except KeyboardInterrupt:
        print(" Shutting down on SIGINT")  # noqa: T201
    finally:
        term_manager.shutdown()
        loop.close()



"""Using a custom thread pool for subprocess writes.
"""
from concurrent import futures

import tornado.web

# This demo requires tornado_xstatic and XStatic-term.js
import tornado_xstatic
from common_demo_stuff import STATIC_DIR, TEMPLATE_DIR, run_and_show_browser

from terminado import SingleTermManager, TermSocket


class TerminalPageHandler(tornado.web.RequestHandler):
    def get(self):
        return self.render(
            "termpage.html",
            static=self.static_url,
            xstatic=self.application.settings["xstatic_url"],
            ws_url_path="/websocket",
        )


def main(argv):
    with futures.ThreadPoolExecutor(max_workers=2) as custom_exec:
        term_manager = SingleTermManager(shell_command=["bash"], blocking_io_executor=custom_exec)
        handlers = [
            (r"/websocket", TermSocket, {"term_manager": term_manager}),
            (r"/", TerminalPageHandler),
            (r"/xstatic/(.*)", tornado_xstatic.XStaticFileHandler, {"allowed_modules": ["termjs"]}),
        ]
        app = tornado.web.Application(
            handlers,
            static_path=STATIC_DIR,
            template_path=TEMPLATE_DIR,
            xstatic_url=tornado_xstatic.url_maker("/xstatic/"),
        )
        app.listen(8765, "localhost")
        run_and_show_browser("http://localhost:8765/", term_manager)


if __name__ == "__main__":
    main([])



"""A separate terminal for every websocket opened.
"""
import tornado.web

# This demo requires tornado_xstatic and XStatic-term.js
import tornado_xstatic
from common_demo_stuff import STATIC_DIR, TEMPLATE_DIR, run_and_show_browser

from terminado import TermSocket, UniqueTermManager


class TerminalPageHandler(tornado.web.RequestHandler):
    def get(self):
        return self.render(
            "termpage.html",
            static=self.static_url,
            xstatic=self.application.settings["xstatic_url"],
            ws_url_path="/websocket",
        )


def main(argv):
    term_manager = UniqueTermManager(shell_command=["bash"])
    handlers = [
        (r"/websocket", TermSocket, {"term_manager": term_manager}),
        (r"/", TerminalPageHandler),
        (r"/xstatic/(.*)", tornado_xstatic.XStaticFileHandler, {"allowed_modules": ["termjs"]}),
    ]
    app = tornado.web.Application(
        handlers,
        static_path=STATIC_DIR,
        template_path=TEMPLATE_DIR,
        xstatic_url=tornado_xstatic.url_maker("/xstatic/"),
    )
    app.listen(8765, "localhost")
    run_and_show_browser("http://localhost:8765/", term_manager)


if __name__ == "__main__":
    main([])



"""Terminals served to xterm.js using Tornado websockets"""

# Copyright (c) Jupyter Development Team
# Copyright (c) 2014, Ramalingam Saravanan <sarava@sarava.net>
# Distributed under the terms of the Simplified BSD License.

from ._version import __version__  # noqa: F401
from .management import (
    NamedTermManager,  # noqa: F401
    SingleTermManager,  # noqa: F401
    TermManagerBase,  # noqa: F401
    UniqueTermManager,  # noqa: F401
)
from .websocket import TermSocket  # noqa: F401



"""A Tornado UI module for a terminal backed by terminado.

See the Tornado docs for information on UI modules:
http://www.tornadoweb.org/en/stable/guide/templates.html#ui-modules
"""
# Copyright (c) Jupyter Development Team
# Copyright (c) 2014, Ramalingam Saravanan <sarava@sarava.net>
# Distributed under the terms of the Simplified BSD License.
from __future__ import annotations

from pathlib import Path

import tornado.web


class Terminal(tornado.web.UIModule):
    """A terminal UI module."""

    def render(self, ws_url: str, cols: int = 80, rows: int = 25) -> str:
        """Render the module."""
        return (
            '<div class="terminado-container" '
            f'data-ws-url="{ws_url}" '
            f'data-rows="{rows}" data-cols="{cols}"/>'
        )

    def javascript_files(self) -> list[str]:
        """Get the list of JS files to include."""
        # TODO: Can we calculate these dynamically?
        return ["/xstatic/termjs/term.js", "/static/terminado.js"]

    def embedded_javascript(self) -> str:
        """Get the embedded JS content as a string."""
        file = Path(__file__).parent / "uimod_embed.js"
        with file.open() as f:
            return f.read()
